/* PCMMIOAD.C    WinSystems PCM-MIO-AD Linux Driver */

/* 
 * $Header$
 * $Id$
 *
 * Filename : $RCSfile$
 * 
 * $Log$
 *
 *
 * WinSystems PCM-MIO-AD Linux Device Driver
 */

 static char *RCSInfo = "$Id: pcmmioad.ko 1.1 2011-08-29 02:28 pdemet Exp $";

/* Portions of original code Copyright (C) 1998-99 by Ori Pomerantz */

/* #define DEBUG 1 */

#ifndef __KERNEL__
#  define __KERNEL__
#endif

#ifndef MODULE
#  define MODULE
#endif

#include <linux/autoconf.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/version.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/delay.h>
#include <linux/mm.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <asm/io.h>
#include <asm/uaccess.h>

// Migration helpers taken from linux kernel source for 2.6.19 so out of tree code will build without errors or warnings
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
	#define IRQF_DISABLED		SA_INTERRUPT
	#define IRQF_SAMPLE_RANDOM	SA_SAMPLE_RANDOM
	#define IRQF_SHARED		SA_SHIRQ
	#define IRQF_PROBE_SHARED	SA_PROBEIRQ
	#define IRQF_PERCPU		SA_PERCPU

	#define IRQF_TRIGGER_LOW	SA_TRIGGER_LOW
	#define IRQF_TRIGGER_HIGH	SA_TRIGGER_HIGH
	#define IRQF_TRIGGER_FALLING	SA_TRIGGER_FALLING
	#define IRQF_TRIGGER_RISING	SA_TRIGGER_RISING
	#define IRQF_TRIGGER_MASK	SA_TRIGGER_MASK
#endif

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Paul DeMetrotion");
MODULE_DESCRIPTION("WinSystems,Inc. PCM-MIO-G-AD Driver");

/* Our own ioctl numbers */

#include "mioad_io.h"

/* Function prototypes for local functions */

int get_buffered_int(int dev_num);
void init_io(unsigned short io_address);
void clr_int(int dev_num, int bit_number);
int get_int(int dev_num);

/* Interrupt handler */
// The structure of the interrupt handler call changed with the 2.6.19 kernel
// this allows builds to succeed for either case

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
   irqreturn_t mio_handler_1(int, void *, struct pt_regs *);
   irqreturn_t mio_handler_2(int, void *, struct pt_regs *);
   irqreturn_t mio_handler_3(int, void *, struct pt_regs *);
   irqreturn_t mio_handler_4(int, void *, struct pt_regs *);
#else
   irqreturn_t mio_handler_1(int, void *);
   irqreturn_t mio_handler_2(int, void *);
   irqreturn_t mio_handler_3(int, void *);
   irqreturn_t mio_handler_4(int, void *);
#endif

// The common interrupt handler

void common_handler(int dev_num);

/* This holds the base addresses of the board */

unsigned short base_port[MAX_DEV] = {0, 0, 0, 0};

/* The name for our device, as it will appear in 
 * /proc/devices */

#define MAJOR_NUM 120
#define DEVICE_NAME "pcmmioad"

/* The maximum length of the message for the device */

#define BUF_LEN 40

/* The message the device will give when asked */

static char Message[MAX_DEV][BUF_LEN];

/* A message pointer for the device */

int Message_Ptr[MAX_DEV] = {0, 0, 0, 0};

#define MAX_INTS 1024

/* Our insmod command line arguments */

static int arr_argc_io = 0;
static int arr_argc_irq = 0;
static unsigned short io[MAX_DEV] = {0, 0, 0, 0};
static unsigned short irq[MAX_DEV] = {0, 0, 0, 0};

module_param_array(io,ushort,&arr_argc_io,S_IRUGO);
module_param_array(irq,ushort,&arr_argc_irq,S_IRUGO);

/* We will buffer up the transition interrupts and will pass them on
   to waiting applications */

unsigned char int_buffer[MAX_DEV][MAX_INTS];
int inptr[MAX_DEV] = {0, 0, 0, 0};
int outptr[MAX_DEV] = {0, 0, 0, 0};

/* These declarations create the wait queues. Five for each supported device. */

DECLARE_WAIT_QUEUE_HEAD(wq_a2d_1_1);
DECLARE_WAIT_QUEUE_HEAD(wq_a2d_1_2);
DECLARE_WAIT_QUEUE_HEAD(wq_dio_1);
DECLARE_WAIT_QUEUE_HEAD(wq_a2d_2_1);
DECLARE_WAIT_QUEUE_HEAD(wq_a2d_2_2);
DECLARE_WAIT_QUEUE_HEAD(wq_dio_2);
DECLARE_WAIT_QUEUE_HEAD(wq_a2d_3_1);
DECLARE_WAIT_QUEUE_HEAD(wq_a2d_3_2);
DECLARE_WAIT_QUEUE_HEAD(wq_dio_3);
DECLARE_WAIT_QUEUE_HEAD(wq_a2d_4_1);
DECLARE_WAIT_QUEUE_HEAD(wq_a2d_4_2);
DECLARE_WAIT_QUEUE_HEAD(wq_dio_4);

unsigned char dac2_port_image[MAX_DEV] = {0, 0, 0, 0};

/***********************************************************************
*			Common Interrupt Handler
*
*       Called by the actual hardware ISR
************************************************************************/

void common_handler(int dev_num)
{
	unsigned char status, int_num;

	/* Read the interrupt ID register from ADC2 */
	dac2_port_image[dev_num] = dac2_port_image[dev_num] | 0x20;
	outb(dac2_port_image[dev_num], base_port[dev_num] + 0x0f);	
	
	while(1) {
		status = inb(base_port[dev_num] + 0x0f);
			
		/* Clear ADC1 interrupt */
		if(status & 1) {
			inb(base_port[dev_num]+1);	/* Clear interrupt */

			/* Wake up anybody waiting for ADC1 */
			switch(dev_num)
			{
				case 0:
					wake_up_interruptible(&wq_a2d_1_1);
					break;

				case 1:
					wake_up_interruptible(&wq_a2d_2_1);
					break;

				case 2:
					wake_up_interruptible(&wq_a2d_3_1);
					break;

				case 3:
					wake_up_interruptible(&wq_a2d_4_1);
					break;
			}
		}
	
		/* Clear ADC2 interrupt */
		if(status & 2) {			
			inb(base_port[dev_num]+5);	/* Clear interrupt */

			/* Wake up anybody waiting for ADC2 */
			switch(dev_num)
			{
				case 0:
					wake_up_interruptible(&wq_a2d_1_2);
					break;

				case 1:
					wake_up_interruptible(&wq_a2d_2_2);
					break;

				case 2:
					wake_up_interruptible(&wq_a2d_3_2);
					break;

				case 3:
					wake_up_interruptible(&wq_a2d_4_2);
					break;
			}
		}

		/* Clear DAC1 interrupt */
		if(status & 4) {
			inb(base_port[dev_num]+9);	/* Clear interrupt */

		}
	
		/* DIO interrupt - Find out which bit */
		if(status & 8) {
			int_num = get_int(dev_num);

			if(int_num)	{
#ifdef DEBUG
	printk("<1>Buffering DIO interrupt on bit %d\n",int_num);
#endif

				/* Buffer the interrupt */
				int_buffer[dev_num][inptr[dev_num]++] = int_num;
				
				/* Pointer wrap */
				if(inptr[dev_num] == MAX_INTS)
					inptr[dev_num] = 0;
				
				/* Clear the interrupt */
				clr_int(dev_num, int_num);
			}

			/* Wake up anybody waiting for a DIO interrupt */
			switch(dev_num)
			{
				case 0:
					wake_up_interruptible(&wq_dio_1);
					break;

				case 1:
					wake_up_interruptible(&wq_dio_2);
					break;

				case 2:
					wake_up_interruptible(&wq_dio_3);
					break;

				case 3:
					wake_up_interruptible(&wq_dio_4);
					break;
			}
		}

		/* Clear DAC2 Interrupt */
		if(status & 0x10) {
			inb(base_port[dev_num]+0x0d);	/* Clear interrupt */

		}

		if((status & 0x1F) == 0)
			break;
	}

	/* Reset the access to the interrupt ID register */
	dac2_port_image[dev_num] = dac2_port_image[dev_num] & 0xdf;
	outb(dac2_port_image[dev_num], base_port[dev_num] + 0x0f);
}

/***********************************************************************
*			Actual Hardware Interrupt Handler
*
*       It defers to another functon for handling
************************************************************************/

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)

irqreturn_t mio_handler_1(int pirq, void *dev_id, struct pt_regs *regs)
{

#ifdef DEBUG
	printk("<1>MIO Interrupt received on Device 1\n");
#endif

	common_handler(0);
	return IRQ_HANDLED;
}

irqreturn_t mio_handler_2(int pirq, void *dev_id, struct pt_regs *regs)
{

#ifdef DEBUG
	printk("<1>MIO Interrupt received on Device 2\n");
#endif

	common_handler(1);
	return IRQ_HANDLED;
}

irqreturn_t mio_handler_3(int pirq, void *dev_id, struct pt_regs *regs)
{

#ifdef DEBUG
	printk("<1>MIO Interrupt received on Device 3\n");
#endif

	common_handler(2);
	return IRQ_HANDLED;
}

irqreturn_t mio_handler_4(int pirq, void *dev_id, struct pt_regs *regs)
{

#ifdef DEBUG
	printk("<1>MIO Interrupt received on Device 4\n");
#endif

	common_handler(3);
	return IRQ_HANDLED;
}

#else

irqreturn_t mio_handler_1(int irq, void *dev_id)
{

#ifdef DEBUG
	printk("<1>MIO Interrupt received on Device 1\n");
#endif

	common_handler(0);
	return IRQ_HANDLED;
}

irqreturn_t mio_handler_2(int irq, void *dev_id)
{

#ifdef DEBUG
	printk("<1>MIO Interrupt received on Device 2\n");
#endif

	common_handler(1);
	return IRQ_HANDLED;
}

irqreturn_t mio_handler_3(int irq, void *dev_id)
{

#ifdef DEBUG
	printk("<1>MIO Interrupt received on Device 3\n");
#endif

	common_handler(2);
	return IRQ_HANDLED;
}

irqreturn_t mio_handler_4(int irq, void *dev_id)
{

#ifdef DEBUG
	printk("<1>MIO Interrupt received on Device 4\n");
#endif

	common_handler(3);
	return IRQ_HANDLED;
}

#endif

/***********************************************************************
*			DEVICE OPEN
*
* This function is called whenever a process attempts to open 
* the device file.
************************************************************************/

static int device_open(struct inode *inode, struct file *file)
{
unsigned int minor = MINOR(inode->i_rdev);

	if(base_port[minor] == 0)
	{
		printk("<1>**** OPEN ATTEMPT on uninitialized port *****\n");
		return -1;
	}

#ifdef DEBUG
	printk ("device_open(%p)\n", file);
#endif

	/* Rewind the message pointer for this device on a new open.

     This could cause problems for applications that use file I/O
     for talking to the device as we do not limit the number of
     applications that can open the device. The solution is to use
     IOCTL functions to control the device rather than file I/O
     calls.
	*/

	Message_Ptr[minor] = 0;

	return(MIO_SUCCESS);
}

/*************************************************************************
*
*			DEVICE CLOSE
*
**************************************************************************/

int device_release(struct inode *inode, struct file *file)
{
#ifdef DEBUG
	printk ("device_release(%p,%p)\n", inode, file);
#endif 

	return(MIO_SUCCESS);
}

/****************************************************************************
*
*			DEVICE READ
*
*****************************************************************************/

ssize_t device_read(struct file *file, char *buffer, size_t length, loff_t *offset)
{
unsigned short port;
unsigned int minor = MINOR(file->f_dentry->d_inode->i_rdev);
int bytes_read;
int i;

	/* We only allow reading of the 32 valid ports in the device. Beyond
	   that we return EOF. */
	if(Message_Ptr[minor] == 32)
		return 0;	/* EOF Condition */

	port = base_port[minor];

	// Do a fresh read of the ports */
	for(i=0; i<32; i++)
		Message[minor][i] = inb(port+i);

	/* Number of bytes actually written to the buffer */
	bytes_read = 0;

#ifdef DEBUG
	printk("device_read(%d,%p,%p,%d)\n", minor, file, buffer, length);
#endif

	/* Actually put the data into the buffer */
	while (length && (Message_Ptr[minor] < 32))
	{
    /* Because the buffer is in the user data segment, 
     * not the kernel data segment, assignment wouldn't 
     * work. Instead, we have to use put_user which 
     * copies data from the kernel data segment to the 
     * user data segment. */

		put_user((Message[minor][Message_Ptr[minor]++]), buffer++);
		length --;
		bytes_read ++;
	}

#ifdef DEBUG
	printk ("Read %d bytes, %d left\n",bytes_read, length);
#endif

	/* Read functions are supposed to return the number 
	 * of bytes actually inserted into the buffer */

	/* By updating file->f_pos we can allow seeking of our Messge_Ptr
	   which will allow random access to the ports using file I/O */

	file->f_pos = (loff_t) Message_Ptr[minor];

	return bytes_read;
}

/****************************************************************************
*
*				DEVICE WRITE
*
*****************************************************************************/

ssize_t device_write(struct file *file, const char *buffer, size_t length, loff_t *offset)
{
unsigned short port;
unsigned int minor = MINOR(file->f_dentry->d_inode->i_rdev);
int i;

	port = base_port[minor];

#ifdef DEBUG
	printk ("device_write(%p,%s,%d)",file, buffer, length);
#endif

	/* Read in the bytes to write, one at a time */
	for(i=0; i<length && i<BUF_LEN; i++)
	{
		get_user(Message[minor][i], buffer+i);

		/* We output each byte to it's appropriate port as it comes in */
		if(Message_Ptr[minor] == 32)
		{
			file->f_pos = (loff_t) Message_Ptr[minor];
			return i;
		}

		outb(Message[minor][i], port + Message_Ptr[minor]++);
	}

	/* As with device_read if we update the file position. Seeking will
	   work as expected. */

	file->f_pos = (loff_t) Message_Ptr[minor];

	/* Again, return the number of input characters used */

	return i;
}

/****************************************************************************
*
*			DEVICE IOCTL
*
*****************************************************************************/

int device_ioctl(struct inode *inode, struct file *file,
     unsigned int ioctl_num, unsigned long ioctl_param)
{
unsigned short word_val;
unsigned char byte_val;
unsigned char offset_val;
unsigned int minor = MINOR(inode->i_rdev);
int i;

#ifdef DEBUG
	printk("PCMMIOAD - IOCTL call IOCTL CODE %04X\n", ioctl_num);
#endif

	/* Switch according to the ioctl called */
	switch (ioctl_num)
	{
		case	WRITE_ADC_COMMAND:

			byte_val = ioctl_param & 0xff;		/* This is the ADC number */
			offset_val = ioctl_param >> 8;		/* This is the data value */

			if (byte_val)		/* ADC 2 */
				outb(offset_val, base_port[minor] + 0x06);
			else
				outb(offset_val, base_port[minor] + 0x02);

			break;

		case	READ_ADC_DATA:

			byte_val = ioctl_param & 0xff;	/* This is the ADC number */

			if (byte_val)		/* ADC 2 */
				word_val = inw(base_port[minor] + 4);
			else
				word_val = inw(base_port[minor]);

			return word_val;

			break;

		case	READ_ADC_STATUS:

			byte_val = ioctl_param & 0xff;		/* This is the ADC number */
        
			if (byte_val)		/* ADC 2 */
				i = inb(base_port[minor] + 7);
			else
				i = inb(base_port[minor] + 3);

			return i;

			break;

		case	WRITE_DIO_BYTE:

			offset_val = ioctl_param & 0xff;
			byte_val = ioctl_param >> 8;
			outb(byte_val, base_port[minor] + 0x10 + offset_val);

			break;

		case	READ_DIO_BYTE:

			offset_val = ioctl_param & 0xff;
			byte_val = inb(base_port[minor] + 0x10 + offset_val);
			return (byte_val & 0xff);

			break;

		case	MIO_WRITE_REG:

			offset_val = ioctl_param & 0xff;
			byte_val = ioctl_param >> 8;
			outb(byte_val, base_port[minor] + offset_val);

			break;

		case	MIO_READ_REG:

			offset_val = ioctl_param & 0xff;
			byte_val = inb(base_port[minor] + offset_val);
			return (byte_val & 0xff);

			break;

		case	WAIT_A2D_INT_1:

#ifdef DEBUG
	printk("<1>PCMMIOAD - IOCTL wait_ad1_int \n");
	printk("<1>PCMMIOAD : current process %i (%s) going to sleep\n", current->pid,current->comm);
#endif

			switch(minor) {
				case 0:
					interruptible_sleep_on(&wq_a2d_1_1);
					break;

				case 1:
					interruptible_sleep_on(&wq_a2d_2_1);
					break;

				case 2:
					interruptible_sleep_on(&wq_a2d_3_1);
					break;

				case 3:
					interruptible_sleep_on(&wq_a2d_4_1);
					break;
			}

			break;

		case	WAIT_A2D_INT_2:

#ifdef DEBUG
	printk("<1>PCMMIOAD - IOCTL wait_ad2_int \n");
	printk("<1>PCMMIOAD : current process %i (%s) going to sleep\n", current->pid,current->comm);
#endif

			switch(minor) {
				case 0:
					interruptible_sleep_on(&wq_a2d_1_2);
					break;

				case 1:
					interruptible_sleep_on(&wq_a2d_2_2);
					break;

				case 2:
					interruptible_sleep_on(&wq_a2d_3_2);
					break;

				case 3:
					interruptible_sleep_on(&wq_a2d_4_2);
					break;
			}
	
			break;

		case	WAIT_DIO_INT:

			if((i = get_buffered_int(minor)))
				return i;

#ifdef DEBUG
	printk("<1>PCMMIOAD - IOCTL wait_dio_int \n");
	printk("<1>PCMMIOAD : current process %i (%s) going to sleep\n", current->pid,current->comm);
#endif

			switch(minor) {
				case 0:
					interruptible_sleep_on(&wq_dio_1);
					break;

				case 1:
					interruptible_sleep_on(&wq_dio_2);
					break;

				case 2:
					interruptible_sleep_on(&wq_dio_3);
					break;

				case 3:
					interruptible_sleep_on(&wq_dio_4);
					break;
			}

			i = get_buffered_int(minor);

			return i;
	
			break;

		case READ_IRQ_ASSIGNED:

			return(irq[minor] & 0xff);

			break;

		case DIO_GET_INT:

			i = get_buffered_int(minor);

			return(i & 0xff);

			break;
	 }

  return(MIO_SUCCESS);
}

/****************************************************************************
*
*           DEVICE_LSEEK
*
*****************************************************************************/

loff_t device_lseek(struct file *filp, loff_t off, int whence)
{
unsigned short offset;
unsigned int minor = MINOR(filp->f_dentry->d_inode->i_rdev);

	/* Convert the offset value to a more familiar size. After all
	   our device is a fixed size of 10 bytes. */

	offset = (unsigned short) off;

#ifdef DEBUG
	printk("<1>PCMMIOAD:lseek - offset %d mode %d\n",offset,whence);
#endif

	switch(whence)
	{
		case 0:   /* Seek set */
			Message_Ptr[minor] = offset;
			if(Message_Ptr[minor] > 32)
				Message_Ptr[minor] = 32;
			break;

		case 1:   /* Seek from Current */
		    Message_Ptr[minor] += offset;
			if(Message_Ptr[minor] > 32)
				Message_Ptr[minor] = 32;
			break;
	
		default:	  /* Seek from End */
		    return -EINVAL;
			break;
		}

	if(Message_Ptr[minor] < 0)
	{
		Message_Ptr[minor] = 0;
		filp->f_pos = (loff_t) Message_Ptr[minor];
		return -EINVAL;
	}

	if(Message_Ptr[minor] > 32)
	{
	    Message_Ptr[minor] = 32;
	    filp->f_pos = (loff_t) Message_Ptr[minor];
	    return -EINVAL;
	}

	filp->f_pos = (loff_t)Message_Ptr[minor];

	return (loff_t) Message_Ptr[minor];
}

/****************************************************************************
*
*			Module Declarations
*
*****************************************************************************/

/* This structure will hold the functions to be called 
 * when a process does something to the our device */

struct file_operations Fops = {

  owner:	THIS_MODULE,
  llseek:	device_lseek,
  read:		device_read, 
  write:	device_write,
  ioctl:	device_ioctl,
  open:		device_open,
  release:	device_release,
};

/****************************************************************************
*			INIT_MODULE
*
* Initialize the module - Register the character device
*****************************************************************************/

int init_module()
{
int i, ret_val;

	/* Sign-on */
	printk("<1>\nWinSystems PCM-MIO-G-AD Driver. Copyright 2010. All rights reserved.\n");
	printk("<1>%s\n",RCSInfo);
	
	/* Check and Map our I/O region requests */
	for (i=0; i<MAX_DEV; i++)
	{
		if(io[i] != 0)
		{
			if(request_region(io[i],0x20,"PCMMIOAD") == NULL)
			{
				printk("<1>PCMMIOAD : Unable to use I/O Address %04Xh\n",io[i]);
			   	io[i] = 0;
			}
		}
	}

	/* Register the character device */
	ret_val = register_chrdev(MAJOR_NUM, DEVICE_NAME, &Fops);

	/* Negative values signify an error */
	if (ret_val < 0)
	{
   		printk (" %s failed with %d\n",",registering the character device PCMMIOAD ", ret_val);
   		return ret_val;
	}
	
	for (i=0; i<MAX_DEV; i++)
	{
		if(io[i] != 0)
		{
			printk ("<1>PCMMIOAD : Base I/O Address = %04Xh\n",io[i]);
			base_port[i] = io[i];
			init_io(io[i]);
		}

		if((io[i] != 0) && (irq[i] != 0))
		{
			switch(i) {
				case 0:
					if(request_irq(irq[i], mio_handler_1, IRQF_SHARED, "PCMMIOAD", RCSInfo))
						printk("<1>PCMMIOAD : Unable to register IRQ %d\n",irq[i]);
					else
						printk("<1>PCMMIOAD : IRQ = %d\n",irq[i]);
					break;

				case 1:
					if(request_irq(irq[i], mio_handler_2, IRQF_SHARED, "PCMMIOAD", RCSInfo))
						printk("<1>PCMMIOAD : Unable to register IRQ %d\n",irq[i]);
					else
						printk("<1>PCMMIOAD : IRQ = %d\n",irq[i]);
					break;

				case 2:
					if(request_irq(irq[i], mio_handler_3, IRQF_SHARED, "PCMMIOAD", RCSInfo))
						printk("<1>PCMMIOAD : Unable to register IRQ %d\n",irq[i]);
					else
						printk("<1>PCMMIOAD : IRQ = %d\n",irq[i]);
					break;

				case 3:
					if(request_irq(irq[i], mio_handler_4, IRQF_SHARED, "PCMMIOAD", RCSInfo))
						printk("<1>PCMMIOAD : Unable to register IRQ %d\n",irq[i]);
					else
						printk("<1>PCMMIOAD : IRQ = %d\n",irq[i]);
					break;
			}		
		}
	}

	return(MIO_SUCCESS);
}

/***************************************************************************
*			CLEANUP_MODULE
*
* Cleanup - unregister the appropriate file from /proc
****************************************************************************/

void cleanup_module()
{
int i;

	for (i=0; i<MAX_DEV; i++)
	{
		/* Unregister I/O Port usage */
		if(io[i])
		  	release_region(io[i], 0x20);

		if(irq[i])
			free_irq(irq[i], RCSInfo);
	}

	/* Unregister the device */
	unregister_chrdev(MAJOR_NUM, DEVICE_NAME);
}  

/***************************************************************************
*			Start of device subroutines 
*
* This array holds the image values of the last write to each I/O
* port. This allows bit manipulation routines to work without having 
* to actually do a read-modify-write to the I/O port.
***************************************************************************/

unsigned char port_images[6];

void init_io(unsigned short io_address)
{
int x;
unsigned short port;

	/* save the address for later use */
	port = io_address + 0X10;

	/* Clear all of the I/O ports. This also makes them inputs */
	for(x=0; x < 7; x++)
		outb(0,port+x);

	/* Clear the image values as well */
	for(x=0; x < 6; x++)
		port_images[x] = 0;

	/* Set page 2 access, for interrupt enables */
	outb(0x80,port+7);

	/* Clear all interrupt enables */
	outb(0,port+8);
	outb(0,port+9);
	outb(0,port+0x0a);

	/* Restore page 0 register access */
	outb(0,port+7);
}

void clr_int(int dev_num, int bit_number)
{
unsigned short port;
unsigned short temp;
unsigned short mask;
unsigned short dio_port;

	dio_port = base_port[dev_num] + 0x10;

	/* Also adjust bit number */
	--bit_number;

	/* Calculate the I/O address based upon bit number */
	port = (bit_number / 8) + dio_port + 8;

	/* Calculate a bit mask based upon the specified bit number */
	mask = (1 << (bit_number % 8));

	/* Turn on page 2 access */
	outb(0x80, dio_port+7);

	/* Get the current state of the interrupt enable register */
	temp = inb(port);

	/* Temporarily clear only our enable. This clears the interrupt */
	temp= temp & ~mask;    /* Clear the enable for this bit */

	/* Now update the interrupt enable register */
	outb(temp,port);

	/* Re-enable our interrupt bit */
	temp = temp | mask;

	outb(temp,port);

	/* Set access back to page 0 */
	outb(0x00,dio_port+7);
}

int get_int(int dev_num)
{
int temp;
int x;
unsigned short dio_port;

	dio_port = base_port[dev_num] + 0x10;

	/* Read the master interrupt pending register,
           mask off undefined bits */

	temp = inb(dio_port+6) & 0x07;

	/* If there are no pending interrupts, return 0 */
	if((temp & 7) == 0)
	    return 0;

	/* There is something pending, now we need to identify it */

	/* Set access to page 3 for interrupt id register */
	outb(0xc0, dio_port + 7);

	/* Read the interrupt ID register for port 0 */
	temp = inb(dio_port+8);

	/* See if any bit set, if so return the bit number */
	if(temp != 0)
	{
	    for(x=0; x<=7; x++)
	    {
            if(temp & (1 << x))
	     	{
		    outb(0,dio_port+7);
		    return(x+1);
            }
         }
     }

	/* None in port 0, read port 1 interrupt ID register */
	temp = inb(dio_port+9);

	/* See if any bit set, if so return the bit number */
	if(temp != 0)
	{
	    for(x=0; x<=7; x++)
	    {
		if(temp & (1 << x))
		{
		    outb(0,dio_port+7);
		    return(x+9);
		}
	    }
	}

	/* Lastly, read the statur of port 2 interrupt ID register */
	temp = inb(dio_port+0x0a);

	/* If any pending, return the appropriate bit number */
	if(temp != 0)
	{
	    for(x=0; x<=7; x++)
	    {
			if(temp & (1 << x))
			{
			   outb(0,dio_port+7);
			   return(x+17);
			}
	    }
	}

	/* We should never get here unless the hardware is seriously
	   misbehaving, but just to be sure, we'll turn the page access
	   back to 0 and return a 0 for no interrupt found
	*/
	outb(0,dio_port+7);

	return 0;
}

int get_buffered_int(int dev_num)
{
int temp;

	if(irq[dev_num] == 0)
	{
	    temp = get_int(dev_num);
	    if(temp)
			clr_int(dev_num, temp);
	    return temp;
	}

	if(outptr[dev_num] != inptr[dev_num])
	{
	    temp = int_buffer[dev_num][outptr[dev_num]++];
	    if(outptr[dev_num] == MAX_INTS)
			outptr[dev_num] = 0;
	    return temp;
	}

	return 0;
}
